;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
;
; FileName:	serial.asm
;
; Description:	Serial receiver function with test code
;		Performs 7 or 8 bit data receive with synch bit
;		to auto-calibrate with sending unit.
;      
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
;
#define BAUD_RATE		D'19200'
#define BIT_COUNT               D'8'    ; // 8 = 1 sync, 7 data
; 					; // 9 = 1 sync, 8 data
#define SIMULATE_START		0x00
;
; NOTE: !! BAUD_RATE, BIT_COUNT, SIMULATE_START !!
;	  -- must match simulation file (or real world host)
;
#include p12c508A.inc
;
;
;
;  Define the bit values within the control registers
;  (Never refer to data sheets again!)
;
; OPTION register setup
;
#define EN_WU		0x00	; // wakeup on pin change
#define DIS_WU		0x80	;
;				;
#define EN_PU		0x00	; // pullups on GP0, GP1, GP3
#define DIS_PU		0x40	;
;				;
#define T0_INT		0x00    ; // Timer 0 clock is internal
#define T0_EXTERN	0x20	; //  or external
;				;
#define PS_T0		0x00	; // Prescaller used by Timer0
#define PS_WDT		0x08	; //   or the Watch Dog Timer
;				;
#define PS_T0_1		0x08	;
#define PS_T0_2		0x00	; // Timer 0 incremented once after
#define PS_T0_4		0x01	; //   prescaller is incremented
#define PS_T0_8		0x02    ; //   'n' times
#define PS_T0_16	0x03	;
#define PS_T0_32	0x04	;
#define PS_T0_64	0x05	;
#define PS_T0_128	0x06	;
#define PS_T0_256	0x07	;
;				;
#define PS_WDT_1	0x00	; // Watch Dog Timer (WDT)
#define PS_WDT_2	0x01	; //   is incremented once after
#define PS_WDT_4	0x02	; //   prescaller is incremented
#define PS_WDT_8	0x03    ; //   'n' times
#define PS_WDT_16	0x04	;
#define PS_WDT_32	0x05	;
#define PS_WDT_64	0x06	;
#define PS_WDT_128	0x07	;
;				;
;
#ifdef SHIP
	__config	_MCLRE_OFF & _CP_ON  & _WDT_OFF & _IntRC_OSC
#else
	__config	_MCLRE_OFF & _CP_OFF & _WDT_OFF & _IntRC_OSC
#endif
;
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
;------- Constants
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
;
#define TRUE	1
#define FALSE	1
;
#define bit0	0
#define bit1	1
#define bit2	2
#define bit3	3
#define bit4	4
#define bit5	5
#define bit6	6
#define	bit7	7
;
#define mbit0	0x01
#define mbit1	0x02
#define mbit2	0x04
#define mbit3	0x08
#define mbit4	0x10
#define mbit5	0x20
#define mbit6	0x40
#define mbit7	0x80
;
;  Define local variables to 'overlay' temporary variables.  The same
;  restrictions apply as when using overlaid variables between files.
;  Also, note the following:
;  NOTE: none of the local variables can be the same name in this file
;  NOTE: local variable don't show up in map file / symbol list
;
LOCAL1	MACRO	name
#define name	temp1
	endm
;
LOCAL2  MACRO	name
#define name 	temp2
	endm
;
LOCAL3	MACRO	name
#define name	temp3
	endm
;
; Hardware specific constants  
;
;   PIN_TX, PIN_RX - transmit and receive pins connected as follows:
;
#define PIN_TX		GPIO,bit2
#define PIN_RX		GPIO,bit3
;
;   OPTION_xxx - setup on-chip options for serial or timed operation
;
#define OPTION_SERIAL	EN_WU | EN_PU | T0_INT | PS_WDT | PS_T0_1
#define OPTION_TIMER	EN_WU | EN_PU | T0_INT | PS_T0  | PS_256
;
#define TRIS_SETUP	mbit3 | mbit1 | mbit0; //GP0,GP1,GP3 are inputs
;
;  Baud rate vote counter is checked for half of the polling loop
;  samples. The Vote range can vary from 1/2 samples to all samples being
;  counted.   Since counter can vary (based on our in-accurate clock)
;  we subtract 1 from the expected sample value.
;
#define us_SEC			D'1000000'	; // micro-seconds per second
;
#define BIT_WIDTH_uS		(us_SEC / BAUD_RATE) 
;
#define BaudCount		(0x100 - ((( BIT_WIDTH_uS / POLL_SAMPLE_CYCLES)/2) - 1))
;
;  Determine test bit for averaging (assume +/- 10 cycles)
;  90 + 90 = 180 > 128  and 110 + 110 = 220 < 256 --> bit 7
;  40 + 40 =  80 > 64   and  60 +  60 = 120 < 128 --> bit 6
;
#if BAUD_RATE == D'19200'
#define AVG_BIT			bit6
#endif
;
#if BAUD_RATE == D'9600'
#define AVG_BIT			bit7
#endif
;
;  Timer 0 gets stopped for 2 cycles each time we change it.
;
#define TMR0_RELOAD_CCLS	2
;
;  Sample loop is 4 instructions (goto is 2 cycles)
;
#define POLL_SAMPLE_CYCLES	((doneSample - bitSample) + 1)
;
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
;------- Variables
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
;
	UDATA
TMR0_Reload	RES	1	; // calibrated reload value
;				;
ExpectedResult	RES	1	; // used for testing/simulation
;				;
temp1		RES	1	; // used for 'local' variables
temp2		RES	1	; // within any function
temp3		RES	1	;
;
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
;------- Code
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
Reset	CODE	0
;
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
; startup
;
; Description: setup chip and goto "main"
;
; Parameters:	wCalOffset - factory calibration value in w
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
;
startup:				; void startup( wCalOffset )
;					; {
	movwf	OSCCAL			;   OSCCAL = wCalOffset
;					;   // define inputs and outputs
	movlw	TRIS_SETUP		;   TRIS( GPIO ) = TRIS_SETUP;
	tris	GPIO			;
;					;   
	goto	main			;   goto &main();
;					; }
;
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
; Function:	ReceiveByteCal
;		
; Description:	Receives a serial data byte and calibrates TMR0.
;		Calculates the width of the start bit to determine
;		timer reload value. REQUIRES bit 0 to be a '1' 
;		(complement of start bit). Works with 7 or 8 bits.
;
; Notes:
;
; Data Format
;	RS-232 protocol (when inverted at TTL levels):
;	1.  the start bit is a high to low transition
;	2.  bit 0 is the first bit after the start bit
;	3.  the stop bit is high
;
;	-----sssss0000011111222223333344444555556666677777S-----
;	_____      ____                                    _____
;            |____|    |XXXX|xxxx|XXXX|xxxx|XXXX|xxxx|XXXX|
;	     ^    ^+sync bit                     stop bit ^
;            ^start bit     
;
; Sampling:
;	The synch and data bits are sampled for the duration
;	of the calculated bit width.  If half of the samples
;	are high, the resulting bit is also high.  Since we
;	are pre-calculating the 'ideal' number of samples
;	(as opposed to dividing the TMR0_Reload by 5), we must
;	reduce this value to compensate for the inaccuracies
;	of our clock.
;	
;	Bit sampling occurs during the sample loop as follows:
;	____                               
;	   |_________|  bit0   |  bit1   |  bit2  ...
;	   |-|S|S|...|--S|S|...|--S|S|...|--S|S|.....
;	   ^  ^      ^  ^   ^
;	   |  |      |      +> data bit processing, skip 9 cycles
;	   |  |      |  +> data bit sample every 5 cycles                            
;	   |  |      +> sync bit processing, skip 11 cycles
;	   |  +> sample every 3/4 cycles
;	   +> start bit polling, lose 3/4 cycles
;
; Start Bit Timing
;	The initial TMR0 count is the 'bit width', determined after 
;	the synch bit goes high.  We subtract TMR0_RELOAD_CCLS (2) to
;	compensate for re-loading the timer while measuring the
;	start pulse. Since the two sampling loops (waitStart and
;	waitSynch) can both be off by by 3 or 4 cycles, our total
;	inaccuracy for bit width measurement is +1.5 or +3 cycles.
;
;	The TMR0_Reload value is averaged after the first 'timed'
;	value.  This should help us maintain accuracy.
;
; Sync Bit Timing
;	Since the timer is still going, we subtract the actual
;	bit width of the start bit and the bit width of the
;	synch bit (TMR0_RELOAD * 2).  Our processing overhead
;	is thus compensated for when we do the first ADD in the
;	sampling loop. The desired result is to finish sampling
;	the synch bit right on time.
;
; Data Bit Timing
;	By complementing the reload value, we allow the timer
;	to count up to '0' for the duration of the bit pulse
;	width. (Note: Since we are testing bit 7 of TMR0, the
;	bit width must be less than 127 (complemented to 128
;	or higher - for baud rates lower than 9600 (100us@4MHz)
;	the prescaller must be used - this is not tested).  
;	Also, I think we loose one tick on TMR0 with the COMF -
;	if we subtracted this from 0x100, TMR0 would be 1 count
;	higher, but would fail the simulations.
;
;	Each time through the loop, we ADD the reload value to
;	the 'wrapped' value of TMR0, thus compensating for 
;	the time spent processing the last bit received.
;	Since this reload 'costs' us 2 cycles, TMR0 should be
;	slowly getting behind the start of the next bit. However,
;	we will break out of the sampling loop 2 cycles too 
;	soon or two cycles too late. On average, this makes up 
;	for the time lost during the reload (I think (but not 
;	too hard:-)).
;
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
;
ReceiveByteCal			; byte ReceiveByteCal( void )
;				; {
	LOCAL1	rxVote		;        byte rxVote;
	LOCAL2	rxData		;        byte rxData;
	LOCAL3	rxCounter	;        byte rxCounter;
;				;
	movlw	BIT_COUNT	;   rxCounter = BIT_COUNT;
	movwf	rxCounter	;
;				;
	movlw	TMR0_RELOAD_CCLS;   wTemp = TMR0_RELOAD_CCLS;
;				;
waitStart			;   do {  // wait for start bit(0)
	btfsc	PIN_RX		;      ; 
	goto	waitStart	;   } while (PIN_RX == 1);
;				;
	clrf	TMR0		;      TMR0 = 0; 
;				;   // TMR0 counts up to 'time'
;				;   // the start bit width
waitSynch:			;   do {
	btfss	PIN_RX		;
	goto	waitSynch	;   } while ( PIN_RX == 0 );
;				;
;				;   // adjust for TMR0 reload time
	subwf	TMR0,w		;      TMR0_Reload += TMR0 - TMR0_RELOAD_CYCLES;
	addwf	TMR0_Reload,f	;      C = 0; // since TMR0 < 128;
;				;
	btfsc	TMR0_Reload,AVG_BIT;   if ( TMR0_Reload[ AVG_BIT ] == 1 )   
	rrf	TMR0_Reload,f	;         TMR0_Reload = TMR0_Reload/2;
;				;
	clrc			;   // Calc offset to reset TMR0
	rlf	TMR0_Reload,f	;      wTemp = ~(TMR0_Reload * 2);
	comf	TMR0_Reload,w	;      // complement since we always 'add' wTemp 
	rrf	TMR0_Reload,f	;      //   to TMR0 for processing overhead
;				;
	comf	TMR0_Reload,f	;   TMR0_Reload = ~TMR0_Reload;
;				;
nextBit:			;   do { // for synch bit and data bits
	addwf	TMR0,f		;     TMR0 += wTemp;
;				;
	clrf	rxVote		;     rxVote = 0;
;				;
bitSample:			;     do {  // for each bit
;				;
	btfsc	PIN_RX		;        if ( PIN_RX == 1 )
	incf	rxVote, f	;           rxVote = rxVote++; 
;				;
	btfsc	TMR0, bit7	;
	goto	bitSample	;     } while (TMR0 < MAX_BIT_TIME );
;				;   
doneSample			;
        movlw	BaudCount	;     C = ( rxVote > BaudCount );
	addwf	rxVote,w	;
;				;
	rrf	rxData,f	;     rxData = (C * 0x80) + (rxData >> 1);
;				;
	movfw	TMR0_Reload	;     wTemp = TMR0_Reload;
;				;
	decfsz	rxCounter,f	;   } while ( --rxCounter > 0 );
	goto	nextBit		; 
;				;
rxDone				;   
#if BIT_COUNT == 8		;   // if (7 data bits)
	clrc			;   //   remove synch bit
	rrf	rxData,f	;      rxData = rxData >> 1;
#endif				;
;				;   // complement for averaging
	comf	TMR0_Reload,f	;   TMR0_Reload = ~TMR0_Reload;
;				;
	retlw	rxData		;   return( &rxData );
;				; }
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
; Function: 	main
;
; Description:	Receive data and verify results (simulated only).
;		Simulation file must be incremental values, must
;		match BAUD_RATE, BIT_COUNT and SIMULATE_START
;		constants
;
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
;
main					; void main( void )
;					; {    
	LOCAL1	mainData		;    byte mainData;
	LOCAL2	lowBits			;    byte lowBits;
;					;
	movlw	OPTION_SERIAL		;   // setup for serial port timing
	option				;      Option = OPTION_SERIAL
;					;
	bsf	PIN_TX			;   PIN_TX = 1; // flag for testing
;					;
	clrf	TMR0_Reload		;   TMR0_Reload = 0; // prevent averaging on first sample
;					;
	movlw	SIMULATE_START		;
	movwf	ExpectedResult		;   ExpectedResult = SIMULATE_START;
;					;
mainLoop				;   do {
;					;      while ( PIN_RX == 1 ) {
	btfss	PIN_RX			;        // verify serial input is
	goto	mainLoop		;        // valid - should be high
;					;      }
	call	ReceiveByteCal		;      mainData = *(ReceiveByte());
	movwf	FSR			;        // three bytes of code saves
	movfw	INDF			;        // 1 global variable byte (RxData)
	movwf	mainData		;
;					;     // verify our results with 
;					;     // the expected value
verifyResult				;
	tstf	ExpectedResult		;     // ignore first sample, no averaging done
	skpz				;     // - only needed if we do
;					;     //   TMR0_Reload averaging - NOTE:
;					;     //   ASSUME SIMULATE_START is zero!
;					;
	xorwf	ExpectedResult,w	;   } while (mainData == expectedResult++);
	incfsz	ExpectedResult,f	;     // incfsz doesn't change 'Z'
;					;     
	skpnz				;
	goto	mainLoop		; 
;					;   // should never get here
hang					;      while (1)
	bcf	PIN_TX			;       PIN_TX = 0; // do something!
	goto	hang			;      
;
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
;       End of File
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
;
	END
;

